<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Sascha Nitsch (grumpydeveloper)
 **/

namespace Federator\Api\FedUsers;

/**
 * handle activitypub outbox requests
 */
class Outbox implements \Federator\Api\FedUsers\FedUsersInterface
{
    /**
     * main instance
     *
     * @var \Federator\Api $main
     */
    private $main;

    /**
     * constructor
     * @param \Federator\Api $main main instance
     */
    public function __construct($main)
    {
        $this->main = $main;
    }

    /**
     * handle get call
     *
     * @param string|null $_user user to fetch outbox for
     * @return string|false response
     */
    public function get($_user)
    {
        if (!isset($_user)) {
            return false;
        }
        $dbh = $this->main->getDatabase();
        $cache = $this->main->getCache();
        $connector = $this->main->getConnector();

        // get user
        $user = \Federator\DIO\User::getUserByName(
            $dbh,
            $_user,
            $connector,
            $cache
        );
        if ($user->id === null) {
            return false;
        }

        // get posts from user
        $outbox = new \Federator\Data\ActivityPub\Common\Outbox();
        $min = $this->main->extractFromURI("min", "");
        $max = $this->main->extractFromURI("max", "");
        $page = $this->main->extractFromURI("page", "");
        if ($page !== "") {
            $items = \Federator\DIO\Posts::getPostsByUser($dbh, $user->id, $connector, $cache, $min, $max);
            $outbox->setItems($items);
        } else {
            $items = [];
        }
        $host = $_SERVER['SERVER_NAME'];
        $id = 'https://' . $host . '/users/' . $_user . '/outbox';
        $outbox->setPartOf($id);
        $outbox->setID($id);
        if ($page !== '') {
            $id .= '?page=' . urlencode($page);
        }
        if ($page === '' || $outbox->count() == 0) {
            $outbox->setFirst($id);
            $outbox->setLast($id . '&min=0');
        }
        if (sizeof($items) > 0) {
            $newestId = $items[0]->getPublished();
            $oldestId = $items[sizeof($items) - 1]->getPublished();
            $outbox->setNext($id . '&max=' . $newestId);
            $outbox->setPrev($id . '&min=' . $oldestId);
        }
        $obj = $outbox->toObject();
        return json_encode($obj, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT); // todo remove pretty
    }

    /**
     * handle post call
     *
     * @param string|null $_user user to add data to outbox
     * @return string|false response
     */
    public function post($_user)
    {
        $_rawInput = file_get_contents('php://input');

        $allHeaders = getallheaders();
        try {
            $this->main->checkSignature($allHeaders);
        } catch (\Federator\Exceptions\PermissionDenied $e) {
            error_log("Outbox::post Signature check failed: " . $e->getMessage());
            http_response_code(401);
            return false;
        }

        $input = is_string($_rawInput) ? json_decode($_rawInput, true) : null;
        $host = $_SERVER['SERVER_NAME'];
        if (!is_array($input)) {
            error_log("Outbox::post Input wasn't of type array");
            return false;
        }

        if (isset($allHeaders['X-Sender']) && $allHeaders['X-Sender'] === $this->main->getConfig()['keys']['headerSenderName']) {
            $outboxActivity = $this->main->getConnector()->jsonToActivity($input);
        } else {
            $outboxActivity = \Federator\Data\ActivityPub\Factory::newActivityFromJson($input);
        }

        if ($outboxActivity === false) {
            error_log("Outbox::post couldn't create outboxActivity");
            return false;
        }

        $sendTo = $outboxActivity->getCC();
        if ($outboxActivity->getType() === 'Undo') {
            $object = $outboxActivity->getObject();
            if ($object !== null) {
                $sendTo = $object->getCC();
            }
        }

        $users = [];

        foreach ($sendTo as $receiver) {
            if ($receiver === '' || !is_string($receiver)) {
                continue;
            }

            if (str_ends_with($receiver, '/followers')) {
                $followers = $this->fetchAllFollowers($receiver, $host);
                if (is_array($followers)) {
                    $users = array_merge($users, $followers);
                }
            }
        }

        if (empty($users)) { // todo remove, debugging for now
            $rootDir = $_SERVER['DOCUMENT_ROOT'] . '../';
            // Save the raw input and parsed JSON to a file for inspection
            file_put_contents(
                $rootDir . 'logs/outbox.log',
                date('Y-m-d H:i:s') . ": ==== POST Outbox Activity ====\n" . json_encode($outboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
                FILE_APPEND
            );
        }
        if ($_user !== false && !in_array($_user, $users, true)) {
            $users[] = $_user;
        }

        foreach ($users as $user) {
            if (!isset($user)) {
                continue;
            }

            $this->postForUser($user, $outboxActivity);
        }

        return json_encode($outboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT);
    }

    /**
     * handle post call for specific user
     *
     * @param string $_user user to add data to outbox
     * @param \Federator\Data\ActivityPub\Common\Activity $outboxActivity the activity that we received
     * @return boolean response
     */
    private function postForUser($_user, $outboxActivity)
    {
        if (isset($_user)) {
            $dbh = $this->main->getDatabase();
            $cache = $this->main->getCache();
            $connector = $this->main->getConnector();

            // get user
            $user = \Federator\DIO\User::getUserByName(
                $dbh,
                $_user,
                $connector,
                $cache
            );
            if ($user->id === null) {
                error_log("Outbox::postForUser couldn't find user: $_user");
                return false;
            }
        }

        $rootDir = $_SERVER['DOCUMENT_ROOT'] . '../';
        // Save the raw input and parsed JSON to a file for inspection
        file_put_contents(
            $rootDir . 'logs/outbox_' . $_user . '.log',
            date('Y-m-d H:i:s') . ": ==== POST " . $_user . " Outbox Activity ====\n" . json_encode($outboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
            FILE_APPEND
        );

        return true;
    }

    /**
     * fetch all followers from url and return the ones that belong to our server
     *
     * @param string $collectionUrl The url of f.e. the posters followers
     * @param string $host our current host-url
     * @return string[] the names of the followers that are hosted on our server
     */
    private static function fetchAllFollowers(string $collectionUrl, string $host): array
    {
        $users = [];

        [$collectionResponse, $collectionInfo] = \Federator\Main::getFromRemote($collectionUrl, ['Accept: application/activity+json']);
        if ($collectionInfo['http_code'] != 200) {
            error_log("Outbox::fetchAllFollowers Failed to fetch follower collection metadata from $collectionUrl");
            return [];
        }

        $collectionData = json_decode($collectionResponse, true);
        $nextPage = $collectionData['first'] ?? $collectionData['current'] ?? null;

        if (!isset($nextPage)) {
            error_log("Outbox::fetchAllFollowers No 'first' or 'current' page in collection at $collectionUrl");
            return [];
        }

        // Loop through all pages
        while ($nextPage) {
            [$pageResponse, $pageInfo] = \Federator\Main::getFromRemote($nextPage, ['Accept: application/activity+json']);
            if ($pageInfo['http_code'] != 200) {
                error_log("Outbox::fetchAllFollowers Failed to fetch follower page at $nextPage");
                break;
            }

            $pageData = json_decode($pageResponse, true);
            $items = $pageData['orderedItems'] ?? $pageData['items'] ?? [];

            foreach ($items as $followerUrl) {
                $parts = parse_url($followerUrl);
                if (!isset($parts['host']) || !str_ends_with($parts['host'], $host)) {
                    continue;
                }

                [$actorResponse, $actorInfo] = \Federator\Main::getFromRemote($followerUrl, ['Accept: application/activity+json']);
                if ($actorInfo['http_code'] != 200) {
                    error_log("Outbox::fetchAllFollowers Failed to fetch actor data for follower: $followerUrl");
                    continue;
                }

                $actorData = json_decode($actorResponse, true);
                if (isset($actorData['preferredUsername'])) {
                    $users[] = $actorData['preferredUsername'];
                }
            }

            $nextPage = $pageData['next'] ?? null;
        }

        return $users;
    }
}
